/*
 *    Copyright IBM Corporation 1985 - All Rights Reserved
 *
 *  Client library for communication with the ITC Window Manager
 *
 * 	James Gosling, 1983
 *
 */

#ifndef FILE
#include <stdio.h>
#endif /* FILE */

#ifndef FONTMAGIC
#include "font.h"
#endif /* FONTMAGIC */

enum GraphicsOperations {
	GR_MOVE,
	GR_LINE,
	GR_GETDIMS,
	GR_GETDIMSR,
	GR_SETFUNC,
	GR_SETTTL,
	GR_SETDIMS,
	GR_MOUSECHANGE,
	GR_SETRAWIN,
	GR_DISABLEINPUT,
	GR_ENABLEINPUT,
	GR_SETMOUSEINTEREST,
	GR_RASTEROP,
	GR_RASTERSMASH,
	GR_DELETEWINDOW,
	GR_DEFINEFONT,
	GR_DEFINEFONTR,
	GR_SELECTFONT,
	GR_SETMOUSEPREFIX,
	GR_MYPGRPIS,
	GR_ADDMENU,
	GR_DISABLENEWLINES,
	GR_BITSTOREGION,
	GR_SENDFONT,
	GR_HEREISFONT,
	GR_SETCURSOR,
	GR_SETSPACESHIM,
	GR_SETCHARSHIM,
	GR_DEFINECOLOR,
	GR_HEREISCOLOR,
	GR_SELECTCOLOR,
	GR_HIDEME,
	GR_EXPOSEME,
	GR_SETPROGRAMNAME,
	GR_SETMOUSEGRID,
	GR_DEFINEREGION,
	GR_SELECTREGION,
	GR_ZAPREGIONS,
	GR_SETCLIPRECTANGLE,
	GR_ACQUIREINPUTFOCUS,
	GR_GIVEUPINPUTFOCUS,
	GR_IHANDLEACQUISITION,
	GR_WRITETOCUTBUFFER,
	GR_READFROMCUTBUFFER,
	GR_HEREISCUTBUFFER,
	GR_ROTATECUTRING,
	GR_SAVEREGION,
	GR_RESTOREREGION,
	GR_FORGETREGION,
	GR_HEREISREGION,
	GR_FILLTRAPEZOID,
	GR_ZOOMFROM,
	GR_SETMENUPREFIX,
	GR_LINKREGION,
	GR_NAMEREGION,
	GR_IAMDONZ,
	GR_STUFFCUTBUFFER,
	GR_ZAPFONT,
        GR_USER,                /* for user out-of-band signalling */
	GR_CONSTRAINMOUSE,
	GR_SETHELPCALL,
	GR_SWAPCUTBUFFERS,
	GR_IAMRINGMASTER,
	GR_APPENDCUTBUFFERS,
	GR_GETNUMWINS,
	GR_GETNUMWINSR,
	GR_NEWREGION,
	GR_NEWREGIONR,
	GR_DESTROYREGION,
	GR_NEWBITMAP,
	GR_NEWBITMAPR,
	GR_SELECTBITMAP,
	GR_DESTROYBITMAP,
	GR_COPYREGION,
	GR_WRITEREGION,
	GR_READREGION,
        GR_FUNCKEY,             /* the next char is from function key table */
        GR_ARC,                 /* Added by Zalman Stern Aug. 1, 1986. */
        GR_GETMOUSE,            /* Added by Zalman Stern Aug. 11, 1986 */
        GR_GETMOUSER,           /* Added by Zalman Stern Aug 11, 1986 */
        GR_REGIONTOBITS,        /* Added by Richard Cohn Dec. 17, 1987 */
	GR_REGIONTOBITSR,	/* Added by Richard Cohn Dec. 17, 1987 */
        GR_FNSAVEREGION,        /* Added by Richard Cohn Dec. 17, 1987 */
        GR_FILLPOLYGON,         /* Added by Curt McDowell Feb. 28, 1989 */
        GR_FILLRECTELLIPSE,     /* Added by Curt McDowell Apr. 15, 1989 */
        GR_FILLARC,             /* Added by Curt McDowell May 5, 1989 */
	GR_VERSION,		/* Added by John Myers June 12, 1989 */
	GR_SETBACKGROUND,	/* Added by John Myers June 12, 1989 */
	GR_BEEP,		/* Added by Chris Newman June 3, 1990 */
};

#define wm_UserOp (GR_USER|0200)    /* the only code with 0200 bit set */
                                    /* that the user is allowed to send */
                                    /* from wm to client, e.g. via menus */

extern int wm_CurrentFunction;

#define wm_SetFunction(f) { \
    register _wm_tmp = (f); \
    if (_wm_tmp != wm_CurrentFunction) \
        wm_GR_send(GR_SETFUNC, (wm_CurrentFunction = _wm_tmp)); \
}

#define wm_MoveTo(x,y) wm_GR_send(GR_MOVE, x, y)
#define wm_DrawTo(x,y) wm_GR_send(GR_LINE, x, y)

#define wm_ArcTo(xend, yend, ccw, a, b, xoff, yoff) \
    wm_GR_send(GR_ARC, xend, yend, ccw, a, b, xoff, yoff)

/*
 * The following permit functions to be expressed as Boolean combinations
 * of the three primitive functions 'source', 'mask', and 'dest'.
 */

#define	f_source		0xCCCC
#define	f_mask			0xF0F0
#define	f_dest			0xAAAA

#define		f_black		0
#define		f_white		1
#define		f_invert	513
#define		f_copy		514
#define		f_BlackOnWhite	515
#define		f_WhiteOnBlack	516

#define wm_MouseInputToken    (0200|GR_MOUSECHANGE)

#define wm_SetTitle(s)         wm_GR_send(GR_SETTTL, s)
#define wm_SetProgramName(s)   wm_GR_send(GR_SETPROGRAMNAME, s)
#define wm_SetHelp(s)          wm_GR_send(GR_SETHELPCALL, s)
#define wm_GetDimensions(width,height) \
                               (putc(0200|GR_GETDIMS,winout), \
                               wm_GR_waitfor (GR_GETDIMSR, 1), \
                               *(width)=wm_uarg[0],*(height)=wm_uarg[1])
#define wm_SetDimensions(minx, maxx, miny, maxy) \
                               wm_GR_send(GR_SETDIMS, minx, maxx, miny, maxy)
#define wm_SawMouse(action,x,y) \
                               (wm_GR_waitfor(GR_MOUSECHANGE,0), \
			       (*action)=wm_uarg[0], \
                               (*x)=wm_uarg[1], \
			       (*y)=wm_uarg[2])
#define wm_SetRawInput() /* obsolete--now a no-op */
#define wm_DisableInput()      putc(0200|GR_DISABLEINPUT,winout)
#define wm_EnableInput()       putc(0200|GR_ENABLEINPUT,winout)
#define wm_RasterOp(sx,sy,dx,dy,w,h) \
                               wm_GR_send(GR_RASTEROP,sx,sy,dx,dy,w,h)
#define wm_RasterSmash(dx,dy,w,h) \
                               wm_GR_send(GR_RASTERSMASH,dx,dy,w,h)
#define wm_ClearWindow()       putc('\014', winout)
#define wm_SetMouseInterest(mask) \
                               wm_GR_send(GR_SETMOUSEINTEREST, mask)
#define wm_AddMenu(s)          wm_GR_send(GR_ADDMENU, s)
#define wm_StdioWindow()       wm_SelectWindow (0)
#define wm_SelectFont(n)       wm_GR_send(GR_SELECTFONT,(CurFont = n)->magic)
#define wm_SetMousePrefix(p)   wm_GR_send(GR_SETMOUSEPREFIX,p)
#define wm_DisableNewlines()   putc(0200|GR_DISABLENEWLINES,winout)
#define wm_BitsToRegion(id,w,h) \
                               wm_GR_send(GR_BITSTOREGION,id,w,h)
#define wm_RegionToBits(id,w,h) \
                               (wm_GR_send(GR_REGIONTOBITS, id), \
			       wm_GR_waitfor(GR_REGIONTOBITSR, 1), \
                               *(w)=wm_uarg[0],*(h)=wm_uarg[1])
#define wm_SetCursor(f,c)      wm_GR_send(GR_SETCURSOR,(f)->magic,c)
#define wm_SetStandardCursor(c) \
                               wm_GR_send(GR_SETCURSOR,-1,c)
#define wm_SetSpaceShim(s)     wm_GR_send(GR_SETSPACESHIM,s)
#define wm_SetCharShim(s)      wm_GR_send(GR_SETCHARSHIM,s)
#define wm_HideMe()            putc(0200|GR_HIDEME,winout)
#define wm_ExposeMe()          putc(0200|GR_EXPOSEME,winout)
#define wm_SetMouseGrid(n)     wm_GR_send(GR_SETMOUSEGRID,n)
#define wm_DefineRegion(id,x,y,w,h) \
                               wm_GR_send(GR_DEFINEREGION,id,x,y,w,h)
#define wm_SelectRegion(id)    wm_GR_send(GR_SELECTREGION,id)
#define wm_ZapRegions()        putc(0200|GR_ZAPREGIONS,winout)
#define wm_SetClipRectangle(x,y,w,h) \
                               wm_GR_send(GR_SETCLIPRECTANGLE,x,y,w,h)
#define wm_AcquireInputFocus() putc(0200|GR_ACQUIREINPUTFOCUS,winout)
#define wm_WriteToCutBuffer()  putc(0200|GR_WRITETOCUTBUFFER,winout)
#define wm_ReadFromCutBuffer(n) \
                               wm_GR_send(GR_READFROMCUTBUFFER, n)
#define wm_StartReadFromCutBuffer(n, len) \
  			       (wm_GR_send(GR_READFROMCUTBUFFER, n), \
			       wm_GR_waitfor(GR_HEREISCUTBUFFER, 1), \
			       *(len)=wm_uarg[0])
/* wm_FinishReadFromCutBuffer() in library */
#define wm_RotateCutRing(n)    wm_GR_send(GR_ROTATECUTRING,n)
#define wm_StuffCutBuffer()    putc(0200|GR_STUFFCUTBUFFER,winout)
#define wm_IAmRingMaster(nbuf,nch) \
                               wm_GR_send(GR_IAMRINGMASTER,nbuf,nch)
#define wm_AppendCutBuffers()  putc(0200|GR_APPENDCUTBUFFERS,winout)
#define wm_SwapCutBuffers()    putc(0200|GR_SWAPCUTBUFFERS,winout)
#define wm_SaveRegion(id,x,y,w,h) \
                               wm_GR_send(GR_SAVEREGION,id,x,y,w,h)
#define wm_FnSaveRegion(id,x,y,w,h) \
                               wm_GR_send(GR_FNSAVEREGION,id,x,y,w,h)
#define wm_RestoreRegion(id,x,y) \
                               wm_GR_send(GR_RESTOREREGION,id,x,y)
#define wm_ForgetRegion(id)    wm_GR_send(GR_FORGETREGION,id)
#define wm_FillTrapezoid(x1,y1,w1,x2,y2,w2,f,c) \
                               wm_GR_send(GR_FILLTRAPEZOID,x1,y1,w1,x2,y2,w2, \
                               ((int)(f)>0?((struct font *)f)->magic:-1),c)
extern void wm_FillRectangle(/*x,y,w,h,f,c*/); /* Not macro: avoids side effect problems */
extern void wm_FillPolygon(/*n,x[],y[],f,c*/);
                               /* RectEllipse: rectangle with ellipse stuff tacked on */
#define wm_FillRectEllipse(cx,cy,w,h,rx,ry,f,c) \
                               wm_GR_send(GR_FILLRECTELLIPSE,cx,cy,w,h,rx,ry, \
                               ((int)(f)>0?((struct font *)f)->magic:-1),c)
#define wm_FillEllipse(cx,cy,rx,ry,f,c) \
                               wm_FillRectEllipse(cx,cy,0,0,rx,ry,f,c)
extern void wm_FillRoundedRect(/*x,y,w,h,rx,ry,f,c*/);
                               /* Arc & PieSlice: Angles in deg. * 64, from 3 o'clock */
                               /* CCW for positive sweep angle, CW for negative */
#define wm_FillArc(pie,x,y,rx,ry,ang,sweep,f,c) /* pie is boolean */ \
                               wm_GR_send(GR_FILLARC,pie,x,y,rx,ry,ang,sweep, \
                               ((int)(f)>0?((struct font *)f)->magic:-1),c)
#define wm_ZoomFrom(x,y,w,h)   wm_GR_send(GR_ZOOMFROM,x,y,w,h)
#define wm_SetMenuPrefix(s)    wm_GR_send(GR_SETMENUPREFIX,s)
#define wm_LinkRegion(newid,oldid) \
                               wm_GR_send(GR_LINKREGION,newid,oldid)
#define wm_NameRegion(id,name) wm_GR_send(GR_NAMEREGION,id,name)
#define wm_IAmDonZ(name)       wm_GR_send(GR_IAMDONZ,name)
#define wm_RefreshFont(n)      wm_GR_send(GR_ZAPFONT,((n)->magic))
#define wm_ConstrainMouse()    putc(0200|GR_CONSTRAINMOUSE,winout)
#define wm_NumWindows(count,max) \
                               (putc(0200|GR_GETNUMWINS,winout), \
			       wm_GR_waitfor (GR_GETNUMWINSR, 1), \
			       *(count)=wm_uarg[0],*(max)=wm_uarg[1])
#define wm_NewRegion()         (putc(0200|GR_NEWREGION,winout), \
                               wm_GR_waitfor(GR_NEWREGIONR, 1), (int) wm_uarg[0])
#define wm_DestroyRegion(id)   wm_GR_send(GR_DESTROYREGION,id)
#define wm_NewBitmap(width, height) \
                               (wm_GR_send(GR_NEWBITMAP,width,height), \
                               wm_GR_waitfor(GR_NEWBITMAPR, 1), (int) wm_uarg[0])
#define wm_SelectBitmap(id)    (wm_GR_send(GR_SELECTBITMAP, id))
#define wm_DestroyBitmap(id)   (wm_GR_send(GR_DESTROYBITMAP, id))
#define wm_CopyRegion(from_id,from_left,from_top,to_id,to_left,to_top,w,h) \
                               (wm_GR_send(GR_COPYREGION, \
                               from_id,from_left,from_top, \
                               to_id,to_left,to_top,w,h))
#define wm_WriteRegion(id, left, top, w, h, s) \
                               (wm_GR_send(GR_WRITEREGION, id, left, top, w, h, s))
#define wm_ReadRegion(id, left, top, s) \
                               (wm_GR_send(GR_READREGION, id, left, top, s))
#define wm_GetMousePos(x, y)   (putc(0200|GR_GETMOUSE,winout), \
                               wm_GR_waitfor (GR_GETMOUSER, 1), \
                               *(x)=wm_uarg[0],*(y)=wm_uarg[1])
#define wm_Version(n)          (wm_GR_send(GR_VERSION,n), \
                               wm_GR_waitfor (GR_VERSION, 1), (int) wm_uarg[0])
#define wm_SetBackground(id,x,y) wm_GR_send(GR_SETBACKGROUND,id,x,y)
#define wm_Beep()		putc(0200|GR_BEEP,winout)

extern short wm_uarg[6];

struct wm_window_aux {
    struct font *CurFont;
    short   nFonts;
    short   nSlots;
    struct font **font;
    unsigned long hostaddress;
};

struct wm_window {
    struct _iobuf   outs,
                    ins;
    char    inb[200],
            outb[2048];
    struct wm_window_aux a;
};

extern struct font *CurFont;
extern struct wm_window *wm_NewWindow(/* host */);
extern struct wm_window *CurrentUserWindow;
extern struct wm_window_aux *CurrentUserWindowParameters;
/*   XXX - infile() and outfile() go away */
#define outfile(w) (&(w)->outs)
#define wm_outfile(w) (&(w)->outs)
#define infile(w) (&(w)->ins)
#define wm_infile(w) (&(w)->ins)
extern FILE *winin, *winout;

#define WMPORT 2000

/* Special keys no longer exist. Therefore, I have removed all support for them.
 * Any references in this section to left or right special keys should be
 * omitted. Zalman Stern June 23, 1987*/

/* Mouse buttons and events.  Events are indicated in a 16-bit bit mask.  The
   orginal version allocated four event types crossed with three buttons for
   12 bits.  The special keys don't really need motion (at least for now), so
   they get two bits each.  This leaves us with zero bits to use for the
   region crossing event.  Fortunately, UpMovement can never happen,
   so we can steal a bit allocated for it.  Yeccchhh!!! */

/* Action types. Movements are when the mouse moves with the corresponding
   key up or down.  Transitions are when the key goes up or down. UpMovement
   can never happen. */

#define UpMovement	0
#define DownTransition	1
#define UpTransition	2
#define DownMovement	3

/* Keys.  First three are on mouse. */

#define LeftButton	2
#define MiddleButton	1
#define RightButton	0

/* This macro combines action types and keys.  It generate bogus values for
   movement actions combined with special keys. */

#define MouseMask(button, action) (1 << (((button) << 2) + (action)))

/* One more MouseInterest event.  It overlays the UpMovement position
   for LeftButton. */

#define RegionCrossing	(MouseMask(LeftButton, UpMovement))

/* When you get a RegionCrossing event, the other bits of the mouse event
   are not events, but instead tell you the state of the mouse buttons at
   the time.  Just check (action&ButtonDown(button)) to see if the button
   is down. */

#define ButtonDown(button)	MouseMask(button, DownTransition)

/* Grab an impossible mouse event to signal whether or not the client
   understands meta keys. */

#define MetaKey (MouseMask(MiddleButton, UpMovement))

/* Meta-X gets sent as wm_MetaInputToken, then 'X' */

#define wm_MetaInputToken (0200|GR_ENABLEINPUT)

extern struct font *wm_FontStruct ();
extern struct font *wm_DefineFont ();

#define wm_AtLeft		    1
#define wm_AtRight		    2
#define wm_BetweenLeftAndRight	    4
#define wm_AtTop		  010
#define wm_AtBottom		  020
#define wm_AtBaseline		  040
#define wm_BetweenTopAndBottom	 0100
#define wm_BetweenTopAndBaseline 0200

extern char *getprofile(/* char *var */);

#define wm_ArrowCursor			'a'
#define wm_GunsightCursor		'g'
#define wm_CrossCursor			'x'
#define wm_HourglassCursor		'H'
#define wm_RightFingerCursor		'f'
#define wm_HorizontalBarsCursor		'h'
#define wm_LowerRightCornerCursor	'l'
#define wm_PaintbrushCursor		'p'
#define wm_UpperLeftCornerCursor	'u'
#define wm_VerticalBarsCursor		'v'
#define wm_DangerousBendCursor		'w'
#define wm_CaretCursor			'|'
#define wm_GrayCaretCursor		'}'
#define wm_WindShieldCursor		'm'

#define wm_FlagRedraw	'R'
#define wm_FlagExpose	'E'
#define wm_FlagKill	'K'
#define wm_FlagHide	'H'
#define wm_FlagFont	'F'

/* High C lies about being an ANSI standard C compiler. */

#if defined(__STDC__) && !defined(__HIGHC__)
#define program(n) char ProgramName[100] = #n ;
#else /* defined(__STDC__) && !defined(__HIGHC__) */
#define program(n) char ProgramName[100] = "n";
#endif /* defined(__STDC__) && !defined(__HIGHC__) */

#define _flsbuf(c, fp) wm_flsbuf(c, fp)
#define fflush(fp) wm_fflush(fp)
